export interface IPublicKey {
  userId: string
  publicKeyArmored: string
  roomId: string
}

export interface IPrivateKey {
  privateKey: any
  passphrase: string
  userId: string
  privateKeyArmored: string
  publicKeyArmored: string
  revocationCertificate: string
  roomId: string
}
// export interface IMetaMessage {
//   whisper?: boolean
//   fromUser: string
// }

export interface IWriteMessage {
  clearString: string
  userId: string
  roomId: string
}

export interface IReadMessage {
  encString: string
  // meta: IMetaMessage
  roomId: string
  authorId: string
  whisper?: boolean
}

export class EndToEnd {
  openpgp: any
  userId: string
  keys: {
    publics: IPublicKey[]
    privates: IPrivateKey[]
  } = {
    publics: [],
    privates: []
  }

  onSend: (obj: IReadMessage & { userId: string }) => Promise<any>

  constructor(opts: {
    userId: string
    openpgp?: any
    onSend?: (obj: IReadMessage & { userId: string }) => Promise<any>
    keys?: {
      publics: IPublicKey[]
      privates: IPrivateKey[]
    }
  }) {
    // if (!opts) throw new Error('no configuration provided')
    this.userId = opts.userId
    if (opts.openpgp) this.openpgp = opts.openpgp
    if (opts.onSend) this.onSend = opts.onSend
    if (opts.keys) this.keys = opts.keys
  }

  addUser(user: IPublicKey) {
    if (!user || !user.publicKeyArmored || !user.userId || !user.roomId) throw new Error('wrong params for adduser')
    this.keys.publics.push(user)
  }

  async whisperToUser(message: IWriteMessage) {
    const msg = await this.composeMessage(message)
    await this.onSend({
      encString: msg.encString,
      roomId: message.roomId,
      userId: message.userId,
      authorId: msg.authorId,
      whisper: true
    })
  }
  async sendToUsers(message: IWriteMessage) {
    const msgs = await this.composeMessagesForUsers(message)
    const key = this.keys.privates.find(f => f.roomId === message.roomId)

    for (const msg of msgs) {
      await this.onSend({
        encString: msg.encString,
        roomId: message.roomId,
        userId: message.userId,
        authorId: key.userId
      })
    }
  }
  async composeMessage(message: IWriteMessage, openpgp?: any): Promise<IReadMessage> {
    const key = this.keys.publics.find(f => f.userId === message.userId)
    if (!key) throw new Error(message.userId + " key's not found")
    const myKey = this.keys.privates.find(f => f.roomId === message.roomId)
    const mgs = { clearString: message.clearString, roomId: message.roomId }
    const encMsg = Object.assign({ openpgp }, mgs)
    const encrypted = await this.encryptString(encMsg)
    const m: IReadMessage = {
      encString: encrypted,
      roomId: message.roomId,
      authorId: myKey.userId
    }
    return m
  }

  async composeMessagesForUsers(opts: { roomId: string; clearString: string }, openpgp?: any) {
    const users = this.keys.publics.filter(f => f.roomId === opts.roomId)

    const messages: IReadMessage[] = []

    for (const user of users) {
      const clearUserMessage: IWriteMessage = {
        roomId: opts.roomId,
        clearString: opts.clearString,
        userId: user.userId
      }
      const encUserMessage = await this.composeMessage(clearUserMessage, openpgp)

      messages.push(encUserMessage)
    }

    return messages
  }

  async readMessage(message: IReadMessage, openpgp?: any) {
    const roomKey = this.keys.privates.find(f => f.roomId === message.roomId)
    if (!roomKey) throw new Error("can't decrypt")
    const opts = {
      openpgp: openpgp || this.openpgp,
      encString: message.encString,
      privateKey: roomKey.privateKey
    }

    return await this.decryptString(opts)
  }

  async decryptString(opts: { encString: string; privateKey: any; openpgp?: any }) {
    return await EndToEnd.decryptString(this.openpgp || opts.openpgp, opts.privateKey, opts.encString)
  }
  async encryptString(opts: {
    clearString: string
    publicKeyArmored?: string
    openpgp?: any
    roomId: string
  }): Promise<string> {
    const key = this.keys.publics.find(f => f.userId === this.userId && f.roomId === opts.roomId)

    if (key) {
      const settings = {
        publicKeyArmored: key.publicKeyArmored,
        clearString: opts.clearString
      }
      return await EndToEnd.encryptString(this.openpgp || opts.openpgp, settings)
    } else if (opts.publicKeyArmored) {
      const settings = {
        publicKeyArmored: opts.publicKeyArmored,
        clearString: opts.clearString
      }

      return await EndToEnd.encryptString(this.openpgp || opts.openpgp, settings)
    } else {
      throw new Error('wrong params')
    }
  }

  getKey(roomId: string) {
    const roomKey = this.keys.privates.find(f => f.roomId === roomId)

    if (roomKey) return roomKey

    return null
  }

  async setKey(opts: { opengp?: any; roomId: string }) {
    const roomKey = this.keys.privates.find(f => f.roomId === opts.roomId)

    if (roomKey) return roomKey
    const passphrase =
      Math.random()
        .toString(36)
        .substring(2, 15) +
      Math.random()
        .toString(36)
        .substring(2, 15)
    const keys = await EndToEnd.regenKey(this.openpgp || opts.opengp, {
      userId: this.userId,
      passphrase: passphrase
    })

    const newRoomKey: IPrivateKey = {
      privateKeyArmored: keys.privateKeyArmored,
      publicKeyArmored: keys.publicKeyArmored,
      revocationCertificate: keys.revocationCertificate,
      privateKey: keys.privateKey,
      roomId: opts.roomId,
      userId: this.userId,
      passphrase: passphrase
    }

    this.keys.publics.push(newRoomKey)
    return newRoomKey
  }

  static async regenKey(openpgp, opts: { userId: string; passphrase?: string }) {
    if (!openpgp) throw new Error('no opengp provided')
    const passphrase =
      opts.passphrase ||
      Math.random()
        .toString(36)
        .substring(2, 15) +
        Math.random()
          .toString(36)
          .substring(2, 15) // protects the private key
    const opgp = await openpgp.generateKey({
      userIds: [{ name: opts.userId }], // you can pass multiple user IDs
      curve: 'ed25519', // ECC curve name
      passphrase: passphrase
    })
    const privateKey = (await openpgp.key.readArmored(opgp.privateKeyArmored)).keys[0]
    await privateKey.decrypt(passphrase)
    const keys = {
      privateKeyArmored: opgp.privateKeyArmored,
      publicKeyArmored: opgp.publicKeyArmored,
      revocationCertificate: opgp.revocationCertificate,
      privateKey: privateKey
    }
    return keys
  }

  static async encryptString(openpgp, opts: { clearString: string; publicKeyArmored: string }) {
    if (!openpgp) throw new Error('no opengp provided')
    const encrypted = await openpgp.encrypt({
      message: openpgp.message.fromText(opts.clearString), // input as Message object
      publicKeys: (await openpgp.key.readArmored(opts.publicKeyArmored)).keys // for encryption
    })
    return encrypted.data
  }

  static async decryptString(openpgp, privateKey, encString: string) {
    if (!openpgp) throw new Error('no opengp provided')
    const { data: decrypted } = await openpgp.decrypt({
      message: await openpgp.message.readArmored(encString), // parse armored message
      privateKeys: [privateKey] // for decryption
    })
    return decrypted
  }
}
